from __future__ import annotations

from typing import TYPE_CHECKING

from framesss.fea.analysis.analysis import Analysis

if TYPE_CHECKING:
    pass


class GriXYAnalysis(Analysis):
    """Subclass of the :class:`Analysis` class for the implementation of the 2D grid model in XY-plane."""

    pass
